@@ -5,8 +5,6 @@ The :mod:`gevent.queue` module implements multi-producer, multi-consumer queues
 that work across greenlets, with the API similar to the classes found in the
 standard :mod:`Queue` and :class:`multiprocessing <multiprocessing.Queue>` modules.
 
-Changed in version 1.0: Queue(0) now means queue of infinite size, not a channel.
-
 The classes in this module implement iterator protocol. Iterating over queue
 means repeatedly calling :meth:`get <Queue.get>` until :meth:`get <Queue.get>` returns ``StopIteration``.
 
@@ -15,9 +13,13 @@ means repeatedly calling :meth:`get <Queue.get>` until :meth:`get <Queue.get>` r
     >>> queue.put(2)
     >>> queue.put(StopIteration)
     >>> for item in queue:
-    ...    print item
+    ...    print(item)
     1
     2
+
+.. versionchanged:: 1.0
+       ``Queue(0)`` now means queue of infinite size, not a channel. A :exc:`DeprecationWarning`
+       will be issued with this argument.
 """
 
 from __future__ import absolute_import
@@ -33,16 +35,38 @@ Full = __queue__.Full
 Empty = __queue__.Empty
 
 from gevent.timeout import Timeout
-from gevent.hub import get_hub, Waiter, getcurrent
+from gevent.hub import get_hub, Waiter, getcurrent, PY3
+from gevent.hub import InvalidSwitchError
 
 
 __all__ = ['Queue', 'PriorityQueue', 'LifoQueue', 'JoinableQueue', 'Channel']
 
 
-class Queue(object):
-    """Create a queue object with a given maximum size.
+def _safe_remove(deq, item):
+    # For when the item may have been removed by
+    # Queue._unlock
+    try:
+        deq.remove(item)
+    except ValueError:
+        pass
 
-    If *maxsize* is less than or equal to zero or ``None``, the queue size is infinite.
+
+class Queue(object):
+    """
+    Create a queue object with a given maximum size.
+
+    If *maxsize* is less than or equal to zero or ``None``, the queue
+    size is infinite.
+
+    .. versionchanged:: 1.1b3
+       Queues now support :func:`len`; it behaves the same as :meth:`qsize`.
+    .. versionchanged:: 1.1b3
+       Multiple greenlets that block on a call to :meth:`put` for a full queue
+       will now be woken up to put their items into the queue in the order in which
+       they arrived. Likewise, multiple greenlets that block on a call to :meth:`get` for
+       an empty queue will now receive items in the order in which they blocked. An
+       implementation quirk under CPython *usually* ensured this was roughly the case
+       previously anyway, but that wasn't the case for PyPy.
     """
 
     def __init__(self, maxsize=None, items=None):
@@ -54,8 +78,20 @@ class Queue(object):
                               DeprecationWarning, stacklevel=2)
         else:
             self.maxsize = maxsize
-        self.getters = set()
-        self.putters = set()
+        # Explicitly maintain order for getters and putters that block
+        # so that callers can consistently rely on getting things out
+        # in the apparent order they went in. This was once required by
+        # imap_unordered. Previously these were set() objects, and the
+        # items put in the set have default hash() and eq() methods;
+        # under CPython, since new objects tend to have increasing
+        # hash values, this tended to roughly maintain order anyway,
+        # but that's not true under PyPy. An alternative to a deque
+        # (to avoid the linear scan of remove()) might be an
+        # OrderedDict, but it's 2.7 only; we don't expect to have so
+        # many waiters that removing an arbitrary element is a
+        # bottleneck, though.
+        self.getters = collections.deque()
+        self.putters = collections.deque()
         self.hub = get_hub()
         self._event_unlock = None
         if items:
@@ -69,6 +105,8 @@ class Queue(object):
         return type(self)(self.maxsize, self.queue)
 
     def _init(self, maxsize, items=None):
+        # FIXME: Why is maxsize unused or even passed?
+        # pylint:disable=unused-argument
         if items:
             self.queue = collections.deque(items)
         else:
@@ -108,6 +146,29 @@ class Queue(object):
         """Return the size of the queue."""
         return len(self.queue)
 
+    def __len__(self):
+        """
+        Return the size of the queue. This is the same as :meth:`qsize`.
+
+        .. versionadded: 1.1b3
+
+            Previously, getting len() of a queue would raise a TypeError.
+        """
+
+        return self.qsize()
+
+    def __bool__(self):
+        """
+        A queue object is always True.
+
+        .. versionadded: 1.1b3
+
+           Now that queues support len(), they need to implement ``__bool__``
+           to return True for backwards compatibility.
+        """
+        return True
+    __nonzero__ = __bool__
+
     def empty(self):
         """Return ``True`` if the queue is empty, ``False`` otherwise."""
         return not self.qsize()
@@ -139,7 +200,7 @@ class Queue(object):
             # We're in the mainloop, so we cannot wait; we can switch to other greenlets though.
             # Check if possible to get a free slot in the queue.
             while self.getters and self.qsize() and self.qsize() >= self.maxsize:
-                getter = self.getters.pop()
+                getter = self.getters.popleft()
                 getter.switch(getter)
             if self.qsize() < self.maxsize:
                 self._put(item)
@@ -147,16 +208,17 @@ class Queue(object):
             raise Full
         elif block:
             waiter = ItemWaiter(item, self)
-            self.putters.add(waiter)
-            timeout = Timeout.start_new(timeout, Full)
+            self.putters.append(waiter)
+            timeout = Timeout._start_new_or_dummy(timeout, Full)
             try:
                 if self.getters:
                     self._schedule_unlock()
                 result = waiter.get()
-                assert result is waiter, "Invalid switch into Queue.put: %r" % (result, )
+                if result is not waiter:
+                    raise InvalidSwitchError("Invalid switch into Queue.put: %r" % (result, ))
             finally:
                 timeout.cancel()
-                self.putters.discard(waiter)
+                _safe_remove(self.putters, waiter)
         else:
             raise Full
 
@@ -168,6 +230,42 @@ class Queue(object):
         """
         self.put(item, False)
 
+    def __get_or_peek(self, method, block, timeout):
+        # Internal helper method. The `method` should be either
+        # self._get when called from self.get() or self._peek when
+        # called from self.peek(). Call this after the initial check
+        # to see if there are items in the queue.
+
+        if self.hub is getcurrent():
+            # special case to make get_nowait() or peek_nowait() runnable in the mainloop greenlet
+            # there are no items in the queue; try to fix the situation by unlocking putters
+            while self.putters:
+                # Note: get() used popleft(), peek used pop(); popleft
+                # is almost certainly correct.
+                self.putters.popleft().put_and_switch()
+                if self.qsize():
+                    return method()
+            raise Empty()
+
+        if not block:
+            # We can't block, we're not the hub, and we have nothing
+            # to return. No choice...
+            raise Empty()
+
+        waiter = Waiter()
+        timeout = Timeout._start_new_or_dummy(timeout, Empty)
+        try:
+            self.getters.append(waiter)
+            if self.putters:
+                self._schedule_unlock()
+            result = waiter.get()
+            if result is not waiter:
+                raise InvalidSwitchError('Invalid switch into Queue.get: %r' % (result, ))
+            return method()
+        finally:
+            timeout.cancel()
+            _safe_remove(self.getters, waiter)
+
     def get(self, block=True, timeout=None):
         """Remove and return an item from the queue.
 
@@ -182,29 +280,8 @@ class Queue(object):
             if self.putters:
                 self._schedule_unlock()
             return self._get()
-        elif self.hub is getcurrent():
-            # special case to make get_nowait() runnable in the mainloop greenlet
-            # there are no items in the queue; try to fix the situation by unlocking putters
-            while self.putters:
-                self.putters.pop().put_and_switch()
-                if self.qsize():
-                    return self._get()
-            raise Empty
-        elif block:
-            waiter = Waiter()
-            timeout = Timeout.start_new(timeout, Empty)
-            try:
-                self.getters.add(waiter)
-                if self.putters:
-                    self._schedule_unlock()
-                result = waiter.get()
-                assert result is waiter, 'Invalid switch into Queue.get: %r' % (result, )
-                return self._get()
-            finally:
-                self.getters.discard(waiter)
-                timeout.cancel()
-        else:
-            raise Empty
+
+        return self.__get_or_peek(self._get, block, timeout)
 
     def get_nowait(self):
         """Remove and return an item from the queue without blocking.
@@ -225,32 +302,17 @@ class Queue(object):
         (*timeout* is ignored in that case).
         """
         if self.qsize():
+            # XXX: Why doesn't this schedule an unlock like get() does?
             return self._peek()
-        elif self.hub is getcurrent():
-            # special case to make peek(False) runnable in the mainloop greenlet
-            # there are no items in the queue; try to fix the situation by unlocking putters
-            while self.putters:
-                self.putters.pop().put_and_switch()
-                if self.qsize():
-                    return self._peek()
-            raise Empty
-        elif block:
-            waiter = Waiter()
-            timeout = Timeout.start_new(timeout, Empty)
-            try:
-                self.getters.add(waiter)
-                if self.putters:
-                    self._schedule_unlock()
-                result = waiter.get()
-                assert result is waiter, 'Invalid switch into Queue.peek: %r' % (result, )
-                return self._peek()
-            finally:
-                self.getters.discard(waiter)
-                timeout.cancel()
-        else:
-            raise Empty
+
+        return self.__get_or_peek(self._peek, block, timeout)
 
     def peek_nowait(self):
+        """Return an item from the queue without blocking.
+
+        Only return an item if one is immediately available. Otherwise
+        raise the :class:`Empty` exception.
+        """
         return self.peek(False)
 
     def _unlock(self):
@@ -259,15 +321,15 @@ class Queue(object):
             if self.putters and (self.maxsize is None or self.qsize() < self.maxsize):
                 repeat = True
                 try:
-                    putter = self.putters.pop()
+                    putter = self.putters.popleft()
                     self._put(putter.item)
-                except:
+                except: # pylint:disable=bare-except
                     putter.throw(*sys.exc_info())
                 else:
                     putter.switch(putter)
             if self.getters and self.qsize():
                 repeat = True
-                getter = self.getters.pop()
+                getter = self.getters.popleft()
                 getter.switch(getter)
             if not repeat:
                 return
@@ -285,6 +347,10 @@ class Queue(object):
             raise result
         return result
 
+    if PY3:
+        __next__ = next
+        del next
+
 
 class ItemWaiter(Waiter):
     __slots__ = ['item', 'queue']
@@ -314,9 +380,11 @@ class PriorityQueue(Queue):
             self.queue = []
 
     def _put(self, item, heappush=heapq.heappush):
+        # pylint:disable=arguments-differ
         heappush(self.queue, item)
 
     def _get(self, heappop=heapq.heappop):
+        # pylint:disable=arguments-differ
         return heappop(self.queue)
 
 
@@ -340,15 +408,34 @@ class LifoQueue(Queue):
 
 
 class JoinableQueue(Queue):
-    '''A subclass of :class:`Queue` that additionally has :meth:`task_done` and :meth:`join` methods.'''
+    """
+    A subclass of :class:`Queue` that additionally has
+    :meth:`task_done` and :meth:`join` methods.
+    """
 
     def __init__(self, maxsize=None, items=None, unfinished_tasks=None):
+        """
+
+        .. versionchanged:: 1.1a1
+           If *unfinished_tasks* is not given, then all the given *items*
+           (if any) will be considered unfinished.
+
+        """
         from gevent.event import Event
         Queue.__init__(self, maxsize, items)
-        self.unfinished_tasks = unfinished_tasks or 0
         self._cond = Event()
         self._cond.set()
 
+        if unfinished_tasks:
+            self.unfinished_tasks = unfinished_tasks
+        elif items:
+            self.unfinished_tasks = len(items)
+        else:
+            self.unfinished_tasks = 0
+
+        if self.unfinished_tasks:
+            self._cond.clear()
+
     def copy(self):
         return type(self)(self.maxsize, self.queue, self.unfinished_tasks)
 
@@ -380,15 +467,24 @@ class JoinableQueue(Queue):
         if self.unfinished_tasks == 0:
             self._cond.set()
 
-    def join(self):
-        '''Block until all items in the queue have been gotten and processed.
+    def join(self, timeout=None):
+        '''
+        Block until all items in the queue have been gotten and processed.
 
         The count of unfinished tasks goes up whenever an item is added to the queue.
         The count goes down whenever a consumer thread calls :meth:`task_done` to indicate
         that the item was retrieved and all work on it is complete. When the count of
         unfinished tasks drops to zero, :meth:`join` unblocks.
+
+        :param float timeout: If not ``None``, then wait no more than this time in seconds
+            for all tasks to finish.
+        :return: ``True`` if all tasks have finished; if ``timeout`` was given and expired before
+            all tasks finished, ``False``.
+
+        .. versionchanged:: 1.1a1
+           Add the *timeout* parameter.
         '''
-        self._cond.wait()
+        return self._cond.wait(timeout=timeout)
 
 
 class Channel(object):
@@ -440,24 +536,19 @@ class Channel(object):
         waiter = Waiter()
         item = (item, waiter)
         self.putters.append(item)
-        timeout = Timeout.start_new(timeout, Full)
+        timeout = Timeout._start_new_or_dummy(timeout, Full)
         try:
             if self.getters:
                 self._schedule_unlock()
             result = waiter.get()
-            assert result is waiter, "Invalid switch into Channel.put: %r" % (result, )
+            if result is not waiter:
+                raise InvalidSwitchError("Invalid switch into Channel.put: %r" % (result, ))
         except:
-            self._discard(item)
+            _safe_remove(self.putters, item)
             raise
         finally:
             timeout.cancel()
 
-    def _discard(self, item):
-        try:
-            self.putters.remove(item)
-        except ValueError:
-            pass
-
     def put_nowait(self, item):
         self.put(item, False)
 
@@ -472,7 +563,7 @@ class Channel(object):
             timeout = 0
 
         waiter = Waiter()
-        timeout = Timeout.start_new(timeout, Empty)
+        timeout = Timeout._start_new_or_dummy(timeout, Empty)
         try:
             self.getters.append(waiter)
             if self.putters:
@@ -506,3 +597,5 @@ class Channel(object):
         if result is StopIteration:
             raise result
         return result
+
+    __next__ = next # py3
